"""Core runtime primitives for Pulka.

This module exists to provide a concrete package for tooling that
expects ``pulka.core`` to be an importable package rather than an
implicit namespace.
"""

__all__ = []
